/*
 * mark_malloc - a malloc tracer, memory leaks detector
 *
 * see http://sed.free.fr/mark_malloc for the documentation.
 * This is public domain.
 */

#include <stdio.h>
#include <stdlib.h>
#include <bfd.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>

void init_mark_malloc(void);
void aam(void);
void bbm(void);
void mark_malloc(void *);

#ifdef _SYSV_V8_
#define libc(x) the_##x
void *(*libc(calloc))(size_t, size_t);
void *(*libc(malloc))(size_t);
void (*libc(free))(void *);
void *(*libc(realloc))(void *, size_t);
#endif /* _SYSV_V8_ */
#ifdef _SYSV_IA32_
#define libc(x) __libc_##x
extern void *libc(calloc)(size_t, size_t);
extern void *libc(malloc)(size_t);
extern void libc(free)(void *);
extern void *libc(realloc)(void *, size_t);
#endif /* _SYSV_IA32_ */


void (*the_exit)(int);
void (*the__exit)(int);

/* some local declarations */
static void insert_and_backtrace(void *);
static void search_and_remove(void *);
static void mark_malloc_output(void);
static void emergency_exit(int);

/* some convenient defines */
#define inc(x) do {							\
		 if ((x) == 0x7fffffff) {				\
		   fprintf(stderr, "mark_malloc: inc overflow\n");	\
		   abort();						\
		 }							\
		 (x)++;							\
	       } while (0)

#define dec(x) do {							\
		 if ((x) == 0) {					\
		   fprintf(stderr, "mark_malloc: dec underflow\n");	\
		   abort();						\
		 }							\
		 (x)--;							\
	       } while (0)

/* our static globals */
struct trace {
  void *address;
  size_t address_size;
  int size;
  int max_size;
  void **bt;
};

static unsigned int do_mark_malloc;
static unsigned int nb_malloc;
static unsigned int nb_free;

#define STATIC_INC	32
static struct trace *malloc_backtrace;
static unsigned int nb_marked;
static unsigned int max_marked;

static unsigned int lib_init=0;
static unsigned int inside_mark_malloc=0;

static char *save_to;
static char *program_file;
static int traces_max = -1;

static void init_lib(void)
{
  char *libc_path;
  void *p;

  /* useless check, but good practice */
  if (lib_init)
    return;

  lib_init = 1;

  inc(inside_mark_malloc);

  libc_path = getenv("MARK_MALLOC_LIBC_PATH");
  if (!libc_path) {
    fprintf(stderr, "mark_malloc error : MARK_MALLOC_LIBC_PATH environment variable not declared\n");
    emergency_exit(1);
  }
  program_file = getenv("MARK_MALLOC_PROGRAM_FILE");
  if (!program_file) {
    fprintf(stderr, "mark_malloc error : MARK_MALLOC_PROGRAM_FILE environment variable not declared\n");
    emergency_exit(1);
  }
  save_to = getenv("MARK_MALLOC_SAVE_TO");
  if (!save_to) {
    fprintf(stderr, "mark_malloc warning : MARK_MALLOC_SAVE_TO environment variable not declared (stderr will be used)\n");
    save_to = "stderr";
  }
  do {
    char *lim = getenv("MARK_MALLOC_TRACES_MAX");
    int ilim;

    if (!lim)
      break;

    /* it is not perfect, but who will give a strange value there ? */
    if (sscanf(lim, "%d", &ilim) != 1 || ilim <0) {
      fprintf(stderr, "mark_malloc warning : MARK_MALLOC_TRACES_MAX environment variable bad value, using unlimited size\n");
      break;
    }

    traces_max = ilim;
  } while(0);

  /* let's init our globals */
  do_mark_malloc = 1;
  nb_malloc = 0;
  nb_free = 0;

  malloc_backtrace = 0;
  nb_marked = 0;
  max_marked = 0;

  /* let's interface with the libc */
  p = dlopen(libc_path, RTLD_LAZY);
  if (!p) {
    fprintf(stderr, "mark_malloc: dlopen error. You surely gave a wrong MARK_MALLOC_LIBC_PATH\n");
    emergency_exit(1);
  }

  the_exit = dlsym(p, "exit");
  if (!the_exit) {
    perror("mark_malloc: dlsym(exit) error");
    emergency_exit(1);
  }

  the__exit = dlsym(p, "_exit");
  if (!the__exit) {
    perror("mark_malloc: dlsym(_exit) error");
    exit(1);
  }

#ifdef _SYSV_V8_
  the_malloc = dlsym(p, "malloc");
  if (!the_malloc) {
    perror("mark_malloc: dlsym(malloc) error");
    exit(1);
  }

  the_calloc = dlsym(p, "calloc");
  if (!the_calloc) {
    perror("mark_malloc: dlsym(calloc) error");
    exit(1);
  }

  the_realloc = dlsym(p, "realloc");
  if (!the_realloc) {
    perror("mark_malloc: dlsym(realloc) error");
    exit(1);
  }

  the_free = dlsym(p, "free");
  if (!the_free) {
    perror("mark_malloc: dlsym(free) error");
    exit(1);
  }
#endif

  dec(inside_mark_malloc);
}

void *malloc(size_t size)
{
  void *p;

  if (!lib_init)
    init_lib();

  if (inside_mark_malloc) {
    return libc(malloc)(size);
  }

  inc(inside_mark_malloc);
  p = libc(malloc)(size);
  if (p) {
    if (do_mark_malloc)
      insert_and_backtrace(p);
  }
  inc(nb_malloc);
  dec(inside_mark_malloc);
  return p;
}

void *realloc(void *ptr, size_t size)
{
  void *p;

  if (!lib_init)
    init_lib();

  if (inside_mark_malloc) {
    return libc(realloc)(ptr, size);
  }

  inc(inside_mark_malloc);
  search_and_remove(ptr);
  p = libc(realloc)(ptr, size);
  if (p) {
    if (do_mark_malloc)
      insert_and_backtrace(p);
  }
  inc(nb_malloc);
  dec(inside_mark_malloc);
  return p;
}

void *calloc(size_t nelem, size_t elsize)
{
  void *p;

  if (!lib_init)
    init_lib();

  if (inside_mark_malloc) {
    return libc(calloc)(nelem, elsize);
  }

  inc(inside_mark_malloc);
  p = libc(calloc)(nelem, elsize);
  if (p) {
    if (do_mark_malloc)
      insert_and_backtrace(p);
  }
  inc(nb_malloc);
  dec(inside_mark_malloc);
  return p;
}

void free(void *p)
{
  if (!lib_init)
    init_lib();

  if (inside_mark_malloc) {
    return libc(free)(p);
  }

  inc(inside_mark_malloc);
  search_and_remove(p);
  libc(free)(p);
  dec(inside_mark_malloc);
}

void exit(int status)
{
  if (!lib_init)
    init_lib();

  if (inside_mark_malloc) {
    the_exit(status);
  }

  inc(inside_mark_malloc);
  mark_malloc_output();
  the_exit(status);
  /* I've got a warning here, don't know how to remove it... Yes this function doesn't return... */

}

void _exit(int status)
{
  if (!lib_init)
    init_lib();

  if (inside_mark_malloc) {
    return the__exit(status);  /* there is no return, but this is the semantic we want to express */
  }

  inc(inside_mark_malloc);
  mark_malloc_output();
  the__exit(status);
  dec(inside_mark_malloc);
}

void init_mark_malloc(void)
{
  inc(inside_mark_malloc);
  if (!lib_init)
    init_lib();
  do_mark_malloc = 0;
  dec(inside_mark_malloc);
}

void aam(void)
{
  inc(inside_mark_malloc);
  inc(do_mark_malloc);
  dec(inside_mark_malloc);
}

void bbm(void)
{
  inc(inside_mark_malloc);
  dec(do_mark_malloc);
  dec(inside_mark_malloc);
}

void mark_malloc(void *p)
{
  inc(inside_mark_malloc);
  insert_and_backtrace(p);
  dec(inside_mark_malloc);
}

static void insert_and_backtrace(void *p)
{
  int i;

  /* if already there, well, there is a little problem, so a warning */
  for (i=0; i<nb_marked; i++)
    if (malloc_backtrace[i].address == p) {
      fprintf(stderr, "mark_malloc: WARNING trying to mark an already marked buffer %p\n", p);
      return;
    }

  /* let's insert */
  if (nb_marked == max_marked) {
    malloc_backtrace = realloc(malloc_backtrace, (max_marked + STATIC_INC) * sizeof(struct trace));
    if (!malloc_backtrace) {
      fprintf(stderr, "mark_malloc: out of memory\n");
      exit(1);
    }
    memset(&malloc_backtrace[nb_marked], 0, STATIC_INC * sizeof(struct trace));

    /* for a real reliability, we should check the following increment */
    max_marked += STATIC_INC;
  }

  malloc_backtrace[nb_marked].address = p;

  /* let's backtrace */
#ifdef _SYSV_V8_
  {
    unsigned int *sp, *oldsp;
    struct trace *t = &malloc_backtrace[nb_marked];

#if 0
    t->bt = 0;
    t->size = 0;
    t->max_size = 0;
#endif

    __asm__ __volatile__ ("st %%o6, [%0]" :: "r" (&sp) : "memory");
    __asm__ __volatile__ ("st %%i6, [%0]" :: "r" (&oldsp) : "memory");
    /* the next line is sparc v8 compliant, use "flushw" under v9 */
    /* we must flush the windows, because we gonna inspect the stack,
     * so all the registers must be saved into it...
     */
    __asm__ __volatile__ ("ta 3");

    while (oldsp) {
      if (t->size == t->max_size) {
	t->bt = realloc(t->bt, (t->max_size + STATIC_INC) * sizeof (void *));
	if (!t->bt) {
	  fprintf(stderr, "mark_malloc: out of memory\n");
	  exit(1);
	}
	t->max_size += STATIC_INC;
      }
      t->bt[t->size++] = (void *)sp[15]; /* it's some kind of magic */

      sp = oldsp;
      oldsp = (unsigned int *)oldsp[14];
    }

  }
#endif /* _SYSV_V8_ */
#ifdef _SYSV_IA32_
  {
    unsigned int *bp;
    struct trace *t = &malloc_backtrace[nb_marked];

    __asm__ __volatile__ ("mov %%ebp, %0" :: "m" (bp) : "memory");

    while (bp) {
      if (t->size == t->max_size) {
	t->bt = realloc(t->bt, (t->max_size + STATIC_INC) * sizeof (void *));
	if (!t->bt) {
	  fprintf(stderr, "mark_malloc: out of memory\n");
	  exit(1);
	}
	t->max_size += STATIC_INC;
      }
      t->bt[t->size++] = (void *)bp[1]; /* it's some kind of magic */

      bp = (unsigned int *)bp[0];
    }

  }
#endif /* _SYSV_IA32_ */

  inc(nb_marked);
}

static void search_and_remove(void *p)
{
  int i;

  for (i = 0; i < nb_marked; i++) {
    if (malloc_backtrace[i].address == p) {
      struct trace temp;

      dec(nb_marked);

      /* we will reuse the buffer, so exchange them */
      temp = malloc_backtrace[i];
      temp.size = 0;
      temp.address = 0;
      malloc_backtrace[i] = malloc_backtrace[nb_marked];
      malloc_backtrace[nb_marked] = temp;

      break;
    }
  }
}

/*------------------------*/
asymbol **symb;

/* reutrn -1 on error, 0 if ok */
int symbol_list_init(bfd *b)
{
  long size;
  long nb;

  if ((bfd_get_file_flags(b) & HAS_SYMS) == 0) {
    fprintf(stderr, "mark_malloc: no symbol table, I give up...\n");
    return -1;
  }

  size = bfd_get_symtab_upper_bound(b);
  if (size < 0) {
    bfd_perror("mark_malloc: bfd_get_symtab_upper_bound");
    return -1;
  }

  symb = (asymbol **)malloc(size);
  if (!symb) {
    fprintf(stderr, "mark_malloc: out of memory\n");
    return -1;
  }

  nb = bfd_canonicalize_symtab(b, symb);
  if (nb < 0) {
    bfd_perror("mark_malloc: bfd_canonicalize_symtab");
    return -1;
  }

  return 0;
}

/* ripped from binutils */
/* These global variables are used to pass information between
   translate_addresses and find_address_in_section.  */

static bfd_vma pc;
static const char *filename;
static const char *functionname;
static unsigned int line;
static boolean found;

static void
find_address_in_section (abfd, section, data)
     bfd *abfd;
     asection *section;
     void *data /*ATTRIBUTE_UNUSED*/;
{
  bfd_vma vma;
  bfd_size_type size;

  if (found)
    return;

  if ((bfd_get_section_flags (abfd, section) & SEC_ALLOC) == 0)
    return;

  vma = bfd_get_section_vma (abfd, section);
  if (pc < vma)
    return;

  size = bfd_get_section_size_before_reloc (section);
  if (pc >= vma + size)
    return;

  found = bfd_find_nearest_line (abfd, section, symb, pc - vma,
                                 &filename, &functionname, &line);
}
/* end of rip */

void print_address_info(FILE *f, bfd *b, void *a)
{
  pc = (unsigned int)a;
  found = 0;
  bfd_map_over_sections (b, find_address_in_section, 0);

  if (found == 0)
    fprintf(f, "??\t??\t0\n");
  else
    fprintf(f, "%s\t%s\t%u\n", functionname ? functionname : "??",
			       filename ? filename : "??", line);
}
/*------------------------*/

static void mark_malloc_output(void)
{
  int i;
  int nb;
  FILE *f;
  bfd *bfd;

  inc(inside_mark_malloc);

  if (!strcmp(save_to, "stdout"))
    f = stdout;
  else if (!strcmp(save_to, "stderr"))
    f = stderr;
  else
    f = fopen(save_to, "w");
  if (!f) {
    fprintf(stderr, "mark_malloc: error opening %s to output mark_malloc stuff, output will go to stderr\n", save_to);
    f = stderr;
  }

  fprintf(f, "nb_marked = %d nb_malloc = %d nb_free = %d nb_malloc-nb_free = %d\n",
	  nb_marked, nb_malloc, nb_free, nb_malloc - nb_free);

  bfd_init();

  bfd = bfd_openr(program_file, 0);
  if (!bfd) {
    bfd_perror("mark_malloc: bfd_open");
    fprintf(stderr, "mark_malloc error : probably you have a wrong MARK_MALLOC_PROGRAM_FILE or your compiled program is \"strange\"\n");
    exit(1);
  }

  if (bfd_check_format(bfd, bfd_object) == false) {
    bfd_perror("mark_malloc: bfd_check_format");
    fprintf(stderr, "mark_malloc error : probably you have a wrong MARK_MALLOC_PROGRAM_FILE or your compiled program is \"strange\"\n");
    exit(1);
  }

  if (symbol_list_init(bfd) == -1) {
    fprintf(stderr, "mark_malloc: error with symbol_list_init");
    fprintf(stderr, "mark_malloc error : probably you have a wrong MARK_MALLOC_PROGRAM_FILE or your compiled program is \"strange\"\n");
    exit(1);
  }

  nb = nb_marked;
  if (traces_max != -1 && traces_max < nb)
    nb = traces_max;

  fprintf(f, "number of unfreed buffer = %d, showing %d traces\n", nb_marked, nb);

  for (i=0; i<nb; i++) {
    int j;

    fprintf(f, "malloc error #%d trace :\n", i);

    for (j = 0; j < malloc_backtrace[i].size; j++)
      print_address_info(f, bfd, malloc_backtrace[i].bt[j]);

    fprintf(f, "-----------------------------------\n");
  }

  dec(inside_mark_malloc);
}


static void emergency_exit(int status)
{
  kill(getpid(), SIGTERM);
}
